import ast
import json
from typing import ClassVar, List

from cognite.client.data_classes import Asset
from pydantic import Field, validator

from cognite.powerops.data_classes.config_model import Configuration
from cognite.powerops.data_classes.reserve_scenario import Auction


class RKOMBidCombinationConfig(Configuration):
    auction: Auction = Field(alias="bid_auction")
    name: str = Field("default", alias="bid_combination_name")
    rkom_bid_config_external_ids: List[str] = Field(alias="bid_rkom_bid_configs")
    parent_external_id: ClassVar[str] = "rkom_bid_combination_configurations"

    @validator("auction", pre=True)
    def to_enum(cls, value):
        return Auction[value] if isinstance(value, str) else value

    @validator("rkom_bid_config_external_ids", pre=True)
    def parse_string(cls, value):
        return [external_id for external_id in ast.literal_eval(value)] if isinstance(value, str) else value

    @property
    def cdf_asset(self) -> Asset:
        sequence_external_id = f"RKOM_bid_combination_configuration_{self.auction.value}_{self.name}"

        return Asset(
            name=sequence_external_id.replace("_", " "),
            description="Defining which RKOM bid methods should be combined (into the total bid form)",
            external_id=sequence_external_id,
            metadata={
                "bid:auction": self.auction.value,
                "bid:combination_name": self.name,
                "bid:rkom_bid_configs": json.dumps(self.rkom_bid_config_external_ids),
            },
            parent_external_id=self.parent_external_id,
        )



from typing import ClassVar

from cognite.client.data_classes import Asset
from pydantic import BaseModel


class RkomMarketConfig(BaseModel):
    external_id: str
    name: str
    timezone: str
    start_of_week: int
    parent_external_id: ClassVar = "market_configurations"

    @property
    def metadata(self) -> dict:
        return {
            "timezone": self.timezone,
            "start_of_week": self.start_of_week,
        }

    @property
    def cdf_asset(self) -> Asset:
        return Asset(
            external_id=self.external_id,
            name=self.name,
            metadata=self.metadata,
            parent_external_id=self.parent_external_id,
            labels=["market"],
        )

    @staticmethod
    def default() -> "RkomMarketConfig":
        return RkomMarketConfig(
            external_id="market_configuration_statnett_rkom_weekly",
            name="RKOM weekly (Statnett)",
            timezone="Europe/Oslo",
            start_of_week=1,
        )



from typing import ClassVar

from cognite.client.data_classes import Asset

from cognite.powerops.data_classes.config_model import Configuration


class MarketConfig(Configuration):
    external_id: str
    name: str
    max_price: float = None
    min_price: float = None
    time_unit: str = None
    timezone: str
    tick_size: float = None
    trade_lot: float = None
    price_steps: int = None
    parent_external_id: ClassVar = "market_configurations"
    price_unit: str = None

    @property
    def metadata(self) -> dict:
        return {
            "min_price": self.min_price,
            "max_price": self.max_price,
            "timezone": self.timezone,
            "time_unit": self.time_unit,
            "tick_size": self.tick_size,
            "trade_lot": self.trade_lot,
            "price_steps": self.price_steps,
            "price_unit": self.price_unit,
        }

    @property
    def cdf_asset(self) -> Asset:
        return Asset(
            external_id=self.external_id,
            name=self.name,
            metadata=self.metadata,
            parent_external_id=self.parent_external_id,
            labels=["market"],
        )


MARKET_CONFIG_NORDPOOL_DAYAHEAD = MarketConfig(
    external_id="market_configuration_nordpool_dayahead",
    name="Nord Pool Day-ahead",
    max_price=4000,
    min_price=-500,
    time_unit="1h",
    timezone="Europe/Oslo",
    tick_size=0.1,
    trade_lot=0.1,
    price_steps=200,
    price_unit="EUR/MWh",
)



import json
from typing import Dict, List, Optional

from cognite.client.data_classes import Asset
from pydantic import Field, validator

from cognite.powerops.data_classes.common import RelativeTime
from cognite.powerops.data_classes.config_model import Configuration


class BenchmarkingConfig(Configuration):
    bid_date: RelativeTime
    shop_start: RelativeTime = Field(alias="shop_starttime")
    shop_end: RelativeTime = Field(alias="shop_endtime")
    production_plan_time_series: Optional[Dict[str, List[str]]]
    market_config_external_id: str = Field(alias="bid_market_config_external_id")
    relevant_shop_objective_metrics: Dict[str, str] = {
        "grand_total": "Grand Total",
        "total": "Total",
        "sum_penalties": "Sum Penalties",
        "major_penalties": "Major Penalties",
        "minor_penalties": "Minor Penalties",
        "load_value": "Load Value",
        "market_sale_buy": "Market Sale Buy",
        "rsv_end_value_relative": "RSV End Value Relative",
        "startup_costs": "Startup Costs",
        "vow_in_transit": "Vow in Transit",
        "sum_feeding_fee": "Sum Feeding Fee",
        "rsv_tactical_penalty": "RSV Tactical Penalty",
        "rsv_end_value": "RSV End Value",
        "bypass_cost": "Bypass Cost",
        "gate_discharge_cost": "Gate Discharge Cost",
        "reserve_violation_penalty": "Reserve Violation Penalty",
        "load_penalty": "Load Penalty",
    }  # Pydantic handles mutable defaults such that this is OK:
    # https://stackoverflow.com/questions/63793662/how-to-give-a-pydantic-list-field-a-default-value/63808835#63808835
    # TODO: Consider adding relationships to bid process config
    #  assets (or remove the optional part that uses those relationships in power-ops-functions)

    @validator("shop_start", "shop_end", "bid_date", pre=True)
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value

    @property
    def metadata(self) -> dict:
        metadata = {
            "bid:date": str(self.bid_date),
            "shop:starttime": str(self.shop_start),
            "shop:endtime": str(self.shop_end),
            "bid:market_config_external_id": self.market_config_external_id,
            "benchmarking_metrics": json.dumps(self.relevant_shop_objective_metrics),
        }
        if self.production_plan_time_series:
            metadata["benchmark:production_plan_time_series"] = json.dumps(
                self.production_plan_time_series, ensure_ascii=False
            )  # ensure_ascii=False to treat Nordic letters properly
        return metadata

    @property
    def cdf_asset(self) -> Asset:
        return Asset(
            external_id="POWEROPS_dayahead_bidding_benchmarking_config",
            name="Benchmarking config DA",
            description="Configuration for benchmarking of day-ahead bidding",
            metadata=self.metadata,
            parent_external_id="benchmarking_configurations",
            labels=["dayahead_bidding_benchmarking_config"],
        )



import random
import time
from typing import Optional

from cognite.client.data_classes import FileMetadata


class ShopRunLog:
    def __init__(self, shop_run_logs: "ShopRunLogs") -> None:
        self._shop_run_logs = shop_run_logs

    def file(self) -> FileMetadata:
        return FileMetadata(external_id="log-123", name="shop_123.log")

    def read(self) -> str:
        return "\n".join(f"log line {i}" for i in range(10))

    def print(self) -> None:
        print(self.read())


class ShopRunLogs:
    def __init__(self, shop_run_result: "ShopRunResult") -> None:
        self._shop_run_result = shop_run_result

    def cplex(self) -> ShopRunLog:
        return ShopRunLog(self)

    def post_run(self) -> ShopRunLog:
        return ShopRunLog(self)

    def shop(self) -> ShopRunLog:
        return ShopRunLog(self)


class ShopRunResult:
    def __init__(self, shop_run: "ShopRun") -> None:
        self._shop_run = shop_run

    @property
    def success(self) -> bool:
        if random.random() > 0.5:
            return True
        return False

    @property
    def error_message(self) -> Optional[str]:
        if not self.success:
            return "(sample) Error: invalid configuration"
        return None

    @property
    def logs(self) -> Optional[ShopRunLogs]:
        return ShopRunLogs(shop_run_result=self)

    @property
    def objective(self) -> Optional[list]:
        raise NotImplementedError


class ShopRun:
    def __init__(self, *, case=None, run_id=None) -> None:
        self.case = case
        self.run_id = run_id

    def is_complete(self) -> bool:
        """
        Query CDF and fetch the status of the run.
        """
        if random.random() > 0.7:
            return True
        return False

    def wait_until_complete(self) -> ShopRunResult:
        while not self.is_complete():
            print("SHOP is still running...")
            time.sleep(3)
        return ShopRunResult(shop_run=self)


class ShopModel:
    def __init__(self) -> None:
        self.model_id = random.randint(1000, 9999)

    def render_yaml(self) -> str:
        return "sintef_shop_model_yaml_representation"

    def update(self):
        raise NotImplementedError


class ShopModelsAPI:
    def __init__(self, shop_api: "ShopAPI"):
        self._shop_api = shop_api

    def list(self) -> list[ShopModel]:
        return ShopModel()

    def retrieve(self, model_id):
        m = ShopModel()
        m.model_id = model_id
        return m


class ShopRunsApi:
    def __init__(self, shop_api: "ShopAPI"):
        self._shop_api = shop_api

    def trigger(self, case) -> ShopRun:
        return ShopRun(case=case)

    def list(self) -> list[ShopRun]:
        raise NotImplementedError

    def retrieve(self, run_id):
        return ShopRun(run_id=run_id)


class ShopAPI:
    def __init__(self, po_client):
        self._po_client = po_client
        self.models = ShopModelsAPI(shop_api=self)
        self.runs = ShopRunsApi(shop_api=self)



from typing import ClassVar, List

from cognite.client.data_classes import Asset, Sequence
from pydantic import BaseModel

from cognite.powerops.data_classes.cdf_resource_collection import (
    BootstrapResourceCollection,
    SequenceContent,
    SequenceRows,
)
from cognite.powerops.utils.cdf_utils import simple_relationship
from cognite.powerops.utils.labels import RelationshipLabels


class BidMatrixGeneratorConfig(BaseModel):
    plant_names: List[str]
    default_method: str  # TODO: Enum?
    default_function_external_id: str
    column_external_ids: ClassVar = [
        "shop_plant",
        "bid_matrix_generation_method",
        "function_external_id",
    ]

    @staticmethod
    def _metadata(price_area: str, bid_process_config_name: str) -> dict:
        # TODO: Rename this from "shop:type" to something without "shop"
        #   (but check if e.g. power-ops-functions uses this)
        return {
            "bid:price_area": f"price_area_{price_area}",
            "shop:type": "bid_matrix_generator_config",
            "bid:bid_process_configuration_name": bid_process_config_name,
        }

    def get_sequence(self, price_area: str, bid_process_config_name: str) -> Sequence:
        column_def = [{"valueType": "STRING", "externalId": external_id} for external_id in self.column_external_ids]
        return Sequence(
            external_id=f"POWEROPS_bid_matrix_generator_config_{bid_process_config_name}",
            name=f"POWEROPS bid matrix generator config {bid_process_config_name}",
            description="Configuration of bid matrix generation method to use for each plant in the price area",
            columns=column_def,
            metadata=self._metadata(price_area, bid_process_config_name),
        )

    def to_sequence_rows(self) -> SequenceRows:
        return SequenceRows(
            rows=[
                (idx, [plant, self.default_method, self.default_function_external_id])
                for idx, plant in enumerate(self.plant_names)
            ],
            columns_external_ids=self.column_external_ids,
        )

    def to_bootstrap_resources(
        self,
        bid_process_config_asset: Asset,
        price_area: str,
        bid_process_config_name: str,
    ) -> BootstrapResourceCollection:
        bootstrap_resource_collection = BootstrapResourceCollection()

        sequence = self.get_sequence(price_area, bid_process_config_name)
        bootstrap_resource_collection.add(sequence)

        sequence_rows = self.to_sequence_rows()
        bootstrap_resource_collection.add(
            SequenceContent(sequence_external_id=sequence.external_id, data=sequence_rows)
        )

        relationship = simple_relationship(
            source=bid_process_config_asset,
            target=sequence,
            label_external_id=RelationshipLabels.BID_MATRIX_GENERATOR_CONFIG_SEQUENCE,
        )
        bootstrap_resource_collection.add(relationship)

        return bootstrap_resource_collection
